from copy import deepcopy
from django.template import Library
from django.db.models import Q
from ..models import MainMenu, FlatMenu
from ..settings import (
    ACTIVE_CLASS, ACTIVE_ANCESTOR_CLASS, DEFAULT_MAIN_MENU_TEMPLATE,
    DEFAULT_FLAT_MENU_TEMPLATE, DEFAULT_SECTION_MENU_TEMPLATE,
    DEFAULT_CHILDREN_MENU_TEMPLATE)


register = Library()

"""
In all menu templates, menu items are always assigned the following attributes
for you to output in the template:

    * `href`: the value to use for the `href` attribute on the <a> element
    * `text`: the text to use for the link
    * `active_class`: a class to be applied to the <li> element (values are
      `ancestor` if the page is an ancestor of the current page or `active` if
      the page is the current_page)
    * `has_children_in_menu`: a boolean value that will let you know if the
      item has children that should be output as a submenu.
"""


@register.simple_tag(takes_context=True)
def main_menu(
    context, show_multiple_levels=True, allow_repeating_parents=True,
    apply_active_classes=True, template=DEFAULT_MAIN_MENU_TEMPLATE
):
    """Render the MainMenu instance for the current site."""
    request = context['request']
    site = request.site
    try:
        menu = site.main_menu
    except MainMenu.DoesNotExist:
        menu = MainMenu.objects.create(site=site)
    ancestor_ids = request.META.get('CURRENT_PAGE_ANCESTOR_IDS', [])

    context.update({
        'apply_active_classes': apply_active_classes,
        'menu_items': tuple(prime_menu_items(
            menu_items=menu.menu_items.for_display(),
            current_page=context.get('self'),
            current_page_ancestor_ids=ancestor_ids,
            current_site=site,
            check_for_children=show_multiple_levels,
            allow_repeating_parents=allow_repeating_parents,
            apply_active_classes=apply_active_classes,
        ))
    })
    t = context.template.engine.get_template(template)
    return t.render(context)


@register.simple_tag(takes_context=True)
def section_menu(
    context, show_section_root=True, show_multiple_levels=True,
    allow_repeating_parents=True, apply_active_classes=True,
    template=DEFAULT_SECTION_MENU_TEMPLATE
):
    """Render a section menu for the current section."""
    request = context['request']
    current_site = request.site
    current_page = context.get('self')
    section_root = request.META.get('CURRENT_SECTION_ROOT')
    ancestor_ids = request.META.get('CURRENT_PAGE_ANCESTOR_IDS', [])

    if section_root:
        menu_items = prime_menu_items(
            menu_items=section_root.get_children().live().in_menu(),
            current_page=current_page,
            current_page_ancestor_ids=ancestor_ids,
            current_site=current_site,
            check_for_children=show_multiple_levels,
            allow_repeating_parents=allow_repeating_parents,
            apply_active_classes=apply_active_classes,
        )

        """
        We want `section_root` to have the same attributes as primed menu
        items, so it can be used in the same way in a template if required.
        """
        setattr(section_root, 'text', section_root.title)
        setattr(section_root, 'href', section_root.relative_url(current_site))

        """
        Before we can work out an `active_class` for `section_root`, we need
        to find out if it's going to be repeated alongside it's children in a
        subnav.
        """
        extra = None
        if (
            allow_repeating_parents and menu_items and
            getattr(section_root, 'repeat_in_subnav', False)
        ):
            """
            The page should be repeated alongside children in the
            subnav, so we create a new item and add it to the existing
            menu_items
            """
            extra = deepcopy(section_root)
            text = section_root.repeated_item_text or section_root.title
            setattr(extra, 'text', text)
            if apply_active_classes and extra.pk == current_page.pk:
                setattr(extra, 'active_class', ACTIVE_CLASS)
            menu_items.insert(0, extra)
        
        """
        Now we know the subnav/repetition situation, we can set the
        `active_class` for `section_root`
        """
        if apply_active_classes:
            active_class = ACTIVE_ANCESTOR_CLASS
            if(
                (extra is None or not hasattr(extra, 'active_class')) and
                section_root.pk == current_page.pk
            ):
                active_class = ACTIVE_CLASS
            setattr(section_root, 'active_class', active_class)
    else:
        menu_items = []

    context.update({
        'section_root': section_root,
        'apply_active_classes': apply_active_classes,
        'show_section_root': show_section_root,
        'menu_items': tuple(menu_items),
    })
    t = context.template.engine.get_template(template)
    return t.render(context)


@register.simple_tag(takes_context=True)
def flat_menu(
    context, handle, show_menu_heading=True, apply_active_classes=False,
    template=DEFAULT_FLAT_MENU_TEMPLATE
):
    """
    Find a FlatMenu for the current site matching the `handle` provided and
    render it.
    """
    request = context['request']
    current_site = request.site
    ancestor_ids = request.META.get('CURRENT_PAGE_ANCESTOR_IDS', [])

    context.update({
        'menu_handle': handle,
        'show_menu_heading': show_menu_heading,
        'apply_active_classes': apply_active_classes,
    })

    try:
        menu = current_site.flat_menus.get(handle__exact=handle)
        context.update({
            'matched_menu': menu,
            'menu_heading': menu.heading,
            'menu_items': tuple(prime_menu_items(
                menu_items=menu.menu_items.for_display(),
                current_page=context.get('self'),
                current_page_ancestor_ids=ancestor_ids,
                current_site=current_site,
                check_for_children=False,
                apply_active_classes=apply_active_classes,
            )),
        })
    except FlatMenu.DoesNotExist:
        context.update({
            'matched_menu': None,
            'menu_heading': '',
            'menu_items': [],
        })
    t = context.template.engine.get_template(template)
    return t.render(context)


@register.simple_tag(takes_context=True)
def children_menu(
    context, menuitem_or_page, stop_at_this_level=False,
    allow_repeating_parents=None, apply_active_classes=None,
    template=DEFAULT_CHILDREN_MENU_TEMPLATE
):
    """
    Retrieve the children menu items for the `menuitem_or_page` provided, and
    render them as a simple ul list
    """
    request = context['request']
    current_site = request.site
    ancestor_ids = request.META.get('CURRENT_PAGE_ANCESTOR_IDS', [])
    current_page = context.get('self')
    try:
        parent_page = menuitem_or_page.link_page.specific
    except AttributeError:
        parent_page = menuitem_or_page

    if apply_active_classes is None:
        apply_active_classes = context.get('apply_active_classes', True)

    menu_items = prime_menu_items(
        menu_items=parent_page.get_children().live().in_menu(),
        current_page=current_page,
        current_page_ancestor_ids=ancestor_ids,
        current_site=current_site,
        check_for_children=not stop_at_this_level,
        allow_repeating_parents=allow_repeating_parents,
        apply_active_classes=apply_active_classes
    )

    if allow_repeating_parents is None:
        allow_repeating_parents = context.get('allow_repeating_parents', True)

    try:
        if allow_repeating_parents and parent_page.repeat_in_subnav:
            extra_item = deepcopy(parent_page)
            href = parent_page.relative_url(current_site)
            text = parent_page.repeated_item_text or parent_page.title
            setattr(extra_item, 'href', href)
            setattr(extra_item, 'text', text)
            if apply_active_classes and parent_page.pk == current_page.pk:
                setattr(extra_item, 'active_class', ACTIVE_CLASS)
            menu_items.insert(0, extra_item)
    except AttributeError:
            pass

    context.update({
        'parent_page': parent_page,
        'menu_items': tuple(menu_items),
        'allow_repeating_parents': allow_repeating_parents,
        'current_template': template,
    })
    t = context.template.engine.get_template(template)
    return t.render(context)


def prime_menu_items(
    menu_items, current_page, current_page_ancestor_ids, current_site,
    check_for_children=False, allow_repeating_parents=True,
    apply_active_classes=True
):
    """
    Prepare a list of menuitem objects or pages for rendering to a menu
    template.
    """
    primed_menu_items = []
    for item in menu_items:

        """
        `MenuItem` and `Page` both have the relative_url method, which we use
        to get a URL relative to the current site root.
        """
        setattr(item, 'href', item.relative_url(current_site))

        try:
            """
            `menu_items` is a list of `MenuItem` objects, that either links
            to a Page, or a custom URL
            """
            page = item.link_page
            menuitem = item
            setattr(item, 'text', item.menu_text)
        except AttributeError:
            """
            `menu_items` is a list of `Page` objects, not `MenuItem` objects
            """
            page = item
            menuitem = None
            setattr(item, 'text', page.title)

        has_children_in_menu = False
        if page:
            """
            If linking to a page, we only want to include this item
            in the resulting list if that page is set to appear in menus.
            """
            if check_for_children and page.depth > 2:
                """
                Working out whether this item should have a sub nav is
                expensive, so we try to do the working out where absolutely
                necessary.
                """
                if ((menuitem and menuitem.allow_subnav) or menuitem is None):
                    has_children_in_menu = (
                        page.get_children().live().in_menu().exists())
                    setattr(item, 'has_children_in_menu', has_children_in_menu)

            """
            Now we know whether this page has a subnav or not, we can look
            for a `repeat_in_subnav` value on the page's specific model
            object, to see if a link to the same page will be repeated as
            the first child. If so, we only want this item to have an
            `ancestor` class at most, as the repeated nav item will be
            given the `active` class.
            """
            page_is_repeated_in_subnav = False
            if allow_repeating_parents and has_children_in_menu:
                page = page.specific
                page_is_repeated_in_subnav = getattr(page, 'repeat_in_subnav',
                                                     False)

            """
            Now we can figure out which class should be added to this item
            """
            if apply_active_classes:
                if current_page and page.pk == current_page.pk:
                    if page_is_repeated_in_subnav:
                        setattr(item, 'active_class', ACTIVE_ANCESTOR_CLASS)
                    else:
                        setattr(item, 'active_class', ACTIVE_CLASS)
                elif page.depth > 2 and page.pk in current_page_ancestor_ids:
                    setattr(item, 'active_class', ACTIVE_ANCESTOR_CLASS)

            primed_menu_items.append(item)

        elif page is None:
            primed_menu_items.append(item)
    return primed_menu_items



from __future__ import absolute_import, unicode_literals

from django import forms
from django.utils.text import capfirst
from django.utils.translation import ugettext as _
from django.shortcuts import get_object_or_404, redirect
from django.core.urlresolvers import reverse
from django.contrib.admin.utils import quote, unquote

from wagtail.wagtailadmin import messages
from wagtail.wagtailcore.models import Site

from wagtailmodeladmin.views import WMABaseView, WMAFormView
from wagtailmodeladmin.helpers import get_url_name
from .models import MainMenu

edit_url_name = get_url_name(MainMenu._meta, 'edit')


class SiteSwitchForm(forms.Form):
    site = forms.ChoiceField(choices=[])

    class Media:
        js = [
            'wagtailmenus/js/site-switcher.js',
        ]

    def __init__(self, current_site, **kwargs):
        initial_data = {'site': self.get_change_url(current_site)}
        super(SiteSwitchForm, self).__init__(initial=initial_data, **kwargs)
        sites = [(self.get_change_url(site), site)
                 for site in Site.objects.all()]
        self.fields['site'].choices = sites

    @classmethod
    def get_change_url(cls, site):
        return reverse(edit_url_name, args=[site.pk])


class MainMenuIndexView(WMABaseView):

    def dispatch(self, request, *args, **kwargs):
        if not self.permission_helper.has_list_permission(request.user):
            raise PermissionDenied
        site = Site.find_for_request(request)
        return redirect(self.get_edit_url(site))


class MainMenuEditView(WMAFormView):
    page_title = _('Editing')
    object_id = None
    instance = None

    def __init__(self, model_admin, object_id):
        super(MainMenuEditView, self).__init__(model_admin)
        self.object_id = unquote(object_id)
        self.pk_safe = quote(self.object_id)
        self.site = get_object_or_404(Site, id=self.object_id)
        self.edit_url = reverse(edit_url_name, args=[self.object_id])
        self.instance = MainMenu.for_site(self.site)
        self.instance.save()

    def get_meta_title(self):
        return _('Editing %s') % self.opts.verbose_name

    def get_page_subtitle(self):
        return capfirst(self.opts.verbose_name)

    def check_action_permitted(self, user):
        return self.permission_helper.can_edit_object(user, self.instance)

    def dispatch(self, request, *args, **kwargs):
        self.site_switcher = None
        if Site.objects.count() > 1:
            self.site_switcher = SiteSwitchForm(self.site)
            if request.GET.get('site', None):
                return redirect(request.GET['site'])
        return super(MainMenuEditView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(MainMenuEditView, self).get_context_data(**kwargs)
        context.update({
            'site': self.site,
            'site_switcher': self.site_switcher,
        })
        return context

    def form_valid(self, form):
        instance = form.save()
        messages.success(
            self.request, _("%s updated.") % capfirst(self.model_name)
        )
        return redirect(self.edit_url)

    def get_error_message(self):
        return _("The %s could not be saved due to errors.") % self.model_name

    def get_template_names(self):
        return ['wagtailmenus/mainmenu_edit.html']



from __future__ import absolute_import, unicode_literals

from django.contrib.auth import get_user_model
from django.test import TestCase
from bs4 import BeautifulSoup


class TestMenuRendering(TestCase):
    fixtures = ['test.json']

    def test_homepage(self):
        """
        Test that homepage (based on `MenuPage`) renders without errors.
        """
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)

    def test_about_us(self):
        """
        Test that 'About us' page (based on `MenuPage`), with 
        `repeat_in_subnav=True`, renders without errors.
        """
        response = self.client.get('/about-us/')
        self.assertEqual(response.status_code, 200)

    def test_meet_the_team(self):
        """
        Test that 'Meet the team' page (based on `Page`), and within a
        section with subnav, renders without errors.
        """
        response = self.client.get('/about-us/meet-the-team/')
        self.assertEqual(response.status_code, 200)

    def test_staff_vacancies(self):
        """
        Test that 'Staff vacancies' page (based on `Page`), with 
        `show_in_menus=False`, and within a section with subnav, renders
        without errors.
        """
        response = self.client.get('/about-us/staff-vacancies/')
        self.assertEqual(response.status_code, 200)

    def test_news_and_events(self):
        """
        Test that 'News and events' page (based on `MenuPage`), with 
        `repeat_in_subnav=False`, and within a section with subnav, renders
        without errors.
        """
        response = self.client.get('/news-and-events/')
        self.assertEqual(response.status_code, 200)

    def test_non_page(self):
        """
        Test that there are no errors when rendering page template without
        the `wagtailmenus.wagtail_hooks.wagtailmenu_params_helper()` method
        having run to add helpful bits to the context.
        """
        response = self.client.get('/custom-url/')
        self.assertEqual(response.status_code, 200)

    def test_main_menu_output_home(self):
        """
        Test that the HTML output by the 'main_menu' tag on the homepage renders as expected.
        """
        response = self.client.get('/')
        # Add assertions to compare rendered HTML against expected HTML
        expected_menu_html = """<div class="collapse navbar-collapse navbar" id="main_menu_output"><ul class="nav navbar-nav"><li class="active"><a href="/">Home</a></li><li class=" dropdown"><a href="/about-us/" class="dropdown-toggle" id="ddtoggle_6" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">About</a><ul class="dropdown-menu" aria-labelledby="ddtoggle_6"><li class=""><a href="/about-us/">Section home</a></li><li class=""><a href="/about-us/meet-the-team/">Meet the team</a></li><li class=""><a href="/about-us/our-heritage/">Our heritage</a></li><li class=""><a href="/about-us/mission-and-values/">Our mission and values</a></li></ul></li><li class=" dropdown"><a href="/news-and-events/" class="dropdown-toggle" id="ddtoggle_14" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">News &amp; events</a><ul class="dropdown-menu" aria-labelledby="ddtoggle_14"><li class=""><a href="/news-and-events/latest-news/">Latest news</a></li><li class=""><a href="/news-and-events/upcoming-events/">Upcoming events</a></li><li class=""><a href="/news-and-events/press/">In the press</a></li></ul></li><li class=""><a href="http://google.co.uk">Google</a></li><li class=""><a href="/contact-us/">Contact us</a></li></ul></div>"""
        soup = BeautifulSoup(response.content, 'html5lib')
        menu_html = soup.find(id='main_menu_output').decode()
        self.assertHTMLEqual(menu_html, expected_menu_html)

    def test_main_menu_output_about_us(self):
        """
        Test that the HTML output by the 'main_menu' tag on the 'About us' page renders as expected.
        """
        response = self.client.get('/about-us/')
        # Add assertions to compare rendered HTML against expected HTML
        expected_menu_html = """<div class="collapse navbar-collapse navbar" id="main_menu_output"><ul class="nav navbar-nav"><li class=""><a href="/">Home</a></li><li class="ancestor dropdown"><a href="/about-us/" class="dropdown-toggle" id="ddtoggle_6" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">About</a><ul class="dropdown-menu" aria-labelledby="ddtoggle_6"><li class="active"><a href="/about-us/">Section home</a></li><li class=""><a href="/about-us/meet-the-team/">Meet the team</a></li><li class=""><a href="/about-us/our-heritage/">Our heritage</a></li><li class=""><a href="/about-us/mission-and-values/">Our mission and values</a></li></ul></li><li class=" dropdown"><a href="/news-and-events/" class="dropdown-toggle" id="ddtoggle_14" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">News &amp; events</a><ul class="dropdown-menu" aria-labelledby="ddtoggle_14"><li class=""><a href="/news-and-events/latest-news/">Latest news</a></li><li class=""><a href="/news-and-events/upcoming-events/">Upcoming events</a></li><li class=""><a href="/news-and-events/press/">In the press</a></li></ul></li><li class=""><a href="http://google.co.uk">Google</a></li><li class=""><a href="/contact-us/">Contact us</a></li></ul></div>"""
        soup = BeautifulSoup(response.content, 'html5lib')
        menu_html = soup.find(id='main_menu_output').decode()
        self.assertHTMLEqual(menu_html, expected_menu_html)

    def test_main_menu_output_meet_the_team(self):
        """
        Test that the HTML output by the 'main_menu' tag on the 'Meet the team' page renders as expected.
        """
        response = self.client.get('/about-us/meet-the-team/')
        # Add assertions to compare rendered HTML against expected HTML
        expected_menu_html = """<div class="collapse navbar-collapse navbar" id="main_menu_output"><ul class="nav navbar-nav"><li class=""><a href="/">Home</a></li><li class="ancestor dropdown"><a href="/about-us/" class="dropdown-toggle" id="ddtoggle_6" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">About</a><ul class="dropdown-menu" aria-labelledby="ddtoggle_6"><li class=""><a href="/about-us/">Section home</a></li><li class="active"><a href="/about-us/meet-the-team/">Meet the team</a></li><li class=""><a href="/about-us/our-heritage/">Our heritage</a></li><li class=""><a href="/about-us/mission-and-values/">Our mission and values</a></li></ul></li><li class=" dropdown"><a href="/news-and-events/" class="dropdown-toggle" id="ddtoggle_14" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">News &amp; events</a><ul class="dropdown-menu" aria-labelledby="ddtoggle_14"><li class=""><a href="/news-and-events/latest-news/">Latest news</a></li><li class=""><a href="/news-and-events/upcoming-events/">Upcoming events</a></li><li class=""><a href="/news-and-events/press/">In the press</a></li></ul></li><li class=""><a href="http://google.co.uk">Google</a></li><li class=""><a href="/contact-us/">Contact us</a></li></ul></div>"""
        soup = BeautifulSoup(response.content, 'html5lib')
        menu_html = soup.find(id='main_menu_output').decode()
        self.assertHTMLEqual(menu_html, expected_menu_html)

    def test_section_menu_output_about_us(self):
        """
        Test that the HTML output by the 'section_menu' tag on the 'About us' page renders as expected.
        """
        response = self.client.get('/about-us/')
        # Add assertions to compare rendered HTML against expected HTML
        expected_menu_html = """<div id="nav-section"><nav class="nav-section" role="navigation"><a href="/about-us/" class="ancestor section_root">About us</a><ul><li class="active"><a href="/about-us/">Section home</a></li><li class=""><a href="/about-us/meet-the-team/">Meet the team</a></li><li class=""><a href="/about-us/our-heritage/">Our heritage</a></li><li class=""><a href="/about-us/mission-and-values/">Our mission and values</a></li></ul></nav></div>"""
        soup = BeautifulSoup(response.content, 'html5lib')
        menu_html = soup.find(id='nav-section').decode()
        self.assertHTMLEqual(menu_html, expected_menu_html)

    def test_section_menu_output_meet_the_team(self):
        """
        Test that the HTML output by the 'section_menu' tag on the 'Meet the team' page renders as expected.
        """
        response = self.client.get('/about-us/meet-the-team/')
        # Add assertions to compare rendered HTML against expected HTML
        expected_menu_html = """<div id="nav-section"><nav class="nav-section" role="navigation"><a href="/about-us/" class="ancestor section_root">About us</a><ul><li class=""><a href="/about-us/">Section home</a></li><li class="active"><a href="/about-us/meet-the-team/">Meet the team</a></li><li class=""><a href="/about-us/our-heritage/">Our heritage</a></li><li class=""><a href="/about-us/mission-and-values/">Our mission and values</a></li></ul></nav></div>"""
        soup = BeautifulSoup(response.content, 'html5lib')
        menu_html = soup.find(id='nav-section').decode()
        self.assertHTMLEqual(menu_html, expected_menu_html)

    def test_contact_flat_menu_output(self):
        """
        Test that the HTML output by the 'flat_menu' tag (when using the handle 'contact') renders as expected.
        """
        response = self.client.get('/')
        # Add assertions to compare rendered HTML against expected HTML
        expected_menu_html = """<div id="nav-contact"><div class="flat-menu contact no_heading"><ul><li class=""><a href="/contact-us/#offices">Call us</a></li><li class=""><a href="#advisor-chat">Chat to an advisor</a></li><li class=""><a href="#request-callback">Request a callback</a></li></ul></div></div>"""
        soup = BeautifulSoup(response.content, 'html5lib')
        menu_html = soup.find(id='nav-contact').decode()
        self.assertHTMLEqual(menu_html, expected_menu_html)

    def test_footer_flat_menu_output(self):
        """
        Test that the HTML output by the 'flat_menu' tag (when using the handle 'footer') renders as expected.
        """
        response = self.client.get('/')
        # Add assertions to compare rendered HTML against expected HTML
        expected_menu_html = """<div id="nav-footer"><div class="flat-menu footer with_heading"><h4>Important links</h4><ul><li class=""><a href="/legal/accessibility/">Accessibility</a></li><li class=""><a href="/legal/privacy-policy/">Privacy policy</a></li><li class=""><a href="/legal/terms-and-conditions/">Terms and conditions</a></li></ul></div></div>"""
        soup = BeautifulSoup(response.content, 'html5lib')
        menu_html = soup.find(id='nav-footer').decode()
        self.assertHTMLEqual(menu_html, expected_menu_html)
